"""
Unit Tests for utils
"""
import os
import unittest

from sp_tool.sharepoint import utils

BASE_DIR = os.path.dirname(os.path.dirname(__file__))


class SharepointUtilsTests(unittest.TestCase):
    """ Unit Tests for sp_tool.sharepoint.utils"""

    def setUp(self):
        """ Setup """

    def test_urlencode(self):
        """ utils.urlencode() """
        tests = [
            ('', ''),
            (' ', '%20'),
            ('$ & < > ? ; # : = , " \' + %',
             '%24%20%26%20%3C%20%3E%20%3F%20%3B%20%23%20%3A%20%3D%20%2C%20%22'
             '%20%27%20%2B%20%25')
        ]
        for source, expected in tests:
            actual = utils.urlencode(source)
            self.assertEqual(actual, expected)


if __name__ == '__main__':
    unittest.main()



""" Simple shared logging """
import logging

NAME = 'sp-tool'
LOG_FORMAT = "%(name)s:%(levelname).1s:%(message)s"

LOG = logging.getLogger(NAME)
LOG.setLevel(logging.INFO)


def initialize_logging(level=logging.INFO, format=LOG_FORMAT, **kwargs):
    """ Initialize logging (should only be done by cli """
    # pylint: disable=W0622
    logging.basicConfig(level=level, format=format, **kwargs)



""" Sharepoint URL Generator """
import functools
import urllib.parse


class SharepointURL:
    """ Represents/generates Sharepoint URLs"""

    def __init__(self, tenant_name, site_name):
        """ Initialize """
        self.tenant_name = tenant_name
        self.site_name = site_name

    @property
    @functools.lru_cache(maxsize=None)
    def site_uri(self):
        """ Full site uri"""
        return f"/sites/{self.site_name}" if self.site_name else "/"

    @property
    @functools.lru_cache(maxsize=None)
    def host(self):
        """ Sharepoint Host Name"""
        return f"{self.tenant_name}.sharepoint.com"

    @property
    @functools.lru_cache(maxsize=None)
    def base(self):
        """ Server Base URL"""
        return f"https://{self.host}"

    @property
    @functools.lru_cache(maxsize=None)
    def site(self):
        """ Site Base URL"""
        return f"{self.base}/{self.site_uri}"

    @property
    @functools.lru_cache(maxsize=None)
    def api_uri(self):
        """ Get api URI"""
        return f"{self.site_uri}/_api"

    @property
    @functools.lru_cache(maxsize=None)
    def api(self):
        """ Get api URL"""
        return f"{self.base}/{self.api_uri}"

    @property
    @functools.lru_cache(maxsize=None)
    def web_uri(self):
        """ Web API URI"""
        return f"{self.api_uri}/Web"

    @property
    @functools.lru_cache(maxsize=None)
    def web(self):
        """ Get web api URL"""
        return f"{self.base}/{self.web_uri}"

    def file_uri(self, folder, filename, suffix=""):
        """ Generate a file URI base """
        folder = urllib.parse.quote(folder or "")
        filename = urllib.parse.quote(filename or "")
        return f"{self.web_uri}/GetFileByServerRelativePath(" \
               f"decodedurl='{self.site_uri}/{folder}/{filename}'" \
               f"){suffix}"

    def file(self, folder, filename, suffix=""):
        """ Generate a file url"""
        uri = self.file_uri(folder, filename, suffix)
        return f"{self.base}{uri}"

    def folder_uri(self, folder, suffix=""):
        """ Generate a folder URL base """
        folder = urllib.parse.quote(folder or "")
        return f"{self.web_uri}/GetFolderByServerRelativePath" \
               f"(decodedurl='{self.site_uri}/{folder}'){suffix}"

    def folder(self, folder, suffix=""):
        """ Generate a folder url"""
        uri = self.folder_uri(folder, suffix)
        return f"{self.base}{uri}"



"""
sp-tool implementation
"""
import os.path
import traceback
from argparse import Namespace
from functools import lru_cache

import yaml

from sp_tool.sharepoint import SharepointConnector
from sp_tool.tool.file_lister import list_files
from sp_tool.tool.logging import LOG

DEFAULT_CONFIG = dict(
    dry_run=False,
    exclude=[],
    exclude_dirs=[],
    include=[],
    include_dirs=[],
    recurse=True,
    source_dir='.',
    base_path="Shared Documents",
    client_id="",
    path="",
    secret="",
    site="",
    tenant=None,
    checkout=False
)


class SharepointTool:
    """ Sharepoint Publishing Tool """

    def __init__(self, *_args, **kwargs):
        """ Initialize """
        opts = {**DEFAULT_CONFIG, **kwargs}
        self.opts = Namespace(**opts)
        if self.opts.tenant is None:
            raise Exception('Must specify tenant')

    @property
    @lru_cache()
    def files(self):
        """ List of files to publish """
        return list_files(self.source_dir, self.opts.recurse,
                          self.opts.include, self.opts.exclude,
                          self.opts.include_dirs, self.opts.exclude_dirs)

    def publish(self):
        """ Publish to Sharepoint """
        LOG.debug("Publishing with options: \n%s",
                  yaml.safe_dump(vars(self.opts), default_flow_style=False))
        if not self.have_files:
            LOG.fatal("ERROR: No files found, unable to continue...")
            return 1

        try:
            if not self.sharepoint_connected:
                LOG.fatal("ERROR: Failed to connect to sharepoint...")
                return 2
            LOG.info("Found files to upload: \n%s",
                     yaml.safe_dump(self.files, default_flow_style=False))
            self._publish()
        except Exception as ex:
            LOG.error("Unknown Error: Failed to publish: %s", ex)
            if self.opts.debug:
                traceback.print_exc(ex, 7)
            return 4
        return 0

    @property
    def have_files(self):
        """ Check if we have files"""
        return bool(self.files)

    @property
    @lru_cache()
    def source_dir(self):
        """ Normalized source directory"""
        source_dir = self.opts.source_dir
        return source_dir

    @property
    @lru_cache()
    def sharepoint(self):
        """ Get sharepoint connection"""
        sharepoint = SharepointConnector(
            tenant=self.opts.tenant,
            client_id=self.opts.client_id,
            client_secret=self.opts.secret,
            site_name=self.opts.site
        )
        return sharepoint

    @property
    def sharepoint_connected(self):
        """ Check if sharepoint connection is working"""
        return self.sharepoint.connected

    def _publish(self):
        """ Publish files """
        files = self.files
        sharepoint = self.sharepoint
        base = self.source_dir
        sp_base = f"{self.opts.base_path}"
        if self.opts.path:
            sp_base = f"{sp_base}/{self.opts.path}"
        no_files = len(files)
        LOG.info("Processing %s file%s", no_files, "s" if no_files != 1 else "")
        LOG.info("Local path is: '%s' Remote path is: '%s'", base, sp_base)
        created_dir = []
        for file in files:
            LOG.debug("Working on '%s'", file)
            rel_file = file[len(base):] if file.startswith(base) else file
            rel_file = rel_file.strip('/')
            rel_dir = os.path.dirname(rel_file)
            sp_folder = f"{sp_base.rstrip('/')}/{rel_dir.lstrip('/')}"
            sp_folder = sp_folder.strip('/')
            sp_file = os.path.basename(rel_file)
            if not self.opts.dry_run:
                LOG.info("Uploading %s to SP as file '%s' in %s", rel_file,
                         sp_file, sp_folder)
                created_dir += sharepoint.add_folder_path(sp_folder,
                                                          created_dir)
                sharepoint.upload_file(file, target_file=sp_file,
                                       folder=sp_folder,
                                       check_out=self.opts.checkout)
            else:
                LOG.info("DRY RUN:: Would have uploaded file '%s' to '%s/%s' "
                         "(with%s checkout)", file, sp_folder, sp_file,
                         "" if self.opts.checkout else "out")



"""
Mock Logger for testing
"""
import logging


class MockLogger:
    """ Mock Logger"""

    def __init__(self):
        """ Initialize"""
        self.messages = []
        self.last_level = None
        self.last_msg = None

    def reset(self):
        """ Reset to defaults"""
        self.messages = []
        self.last_level = None
        self.last_msg = None

    def log(self, level, msg, *params):
        """ Log a message """
        self.last_level = level
        message = msg % params
        self.last_msg = message
        self.messages.append((level, message, msg, params))

    def debug(self, msg, *params):
        """ Log a debug message """
        self.log(logging.DEBUG, msg, *params)

    def info(self, msg, *params):
        """ Log an info message"""
        self.log(logging.INFO, msg, *params)

    @property
    def last_message(self):
        """ Get last formatted message as a level, message tuple """
        if not self.messages:
            return (None, None)
        return self.messages[-1][0:2]

    @property
    def last_message_raw(self):
        """
            Get last unformatted message as aevel, formatstring, params tuple
        """
        if not self.messages:
            return None, None, []
        return self.messages[-1][0], self.messages[-1][2], self.messages[-1][3]



"""
Exceptions for sharepoint api connector
"""


class BaseSharepointException(Exception):
    """ Base Sharepoint Exception """


class APICallFailedSharepointException(Exception):
    """ Base Sharepoint Exception """

    def __init__(self, code, err_msg):
        """ Initializer"""
        super().__init__(f"Error: {code}: {err_msg}")
        self.code = code
        self.err_msg = err_msg



""" File Lister """
import fnmatch
import os
from typing import List

from sp_tool.tool.logging import LOG


def _matches(file, glob):
    """ check if file matches a glob"""
    return fnmatch.fnmatch(file, glob)


def _matches_one_of(file, globs):
    """ check if file one of the provided globs """
    for glob in globs:
        if fnmatch.fnmatch(file, glob):
            return True
    return False


def _filter_files(file_list, include, exclude):
    """ Filter a list of files """
    files = []
    for file in file_list:
        if not _matches_one_of(file, include):
            LOG.debug("File %s does not match any of the include patterns",
                      file)
            continue
        if _matches_one_of(file, exclude):
            LOG.debug("File %s matches an exclude pattern, ignoring", file)
            continue
        LOG.debug("Found file '%s'", file)
        files.append(file)
    return files


def list_files(directory, recurse=False,
               include: List[str] = None, exclude: List[str] = None,
               include_dirs: List[str] = None, exclude_dirs: List[str] = None):
    """ List files in directory with includes or excludes """
    # pylint: disable=too-many-arguments
    files = []
    src_dir = directory.rstrip('/')

    if not include:
        include = ['*']
    if not include_dirs:
        include_dirs = ['*']
    if not exclude:
        exclude = []
    if not exclude_dirs:
        exclude_dirs = []

    LOG.info("Scanning '%s'.Includes: %s. Excludes: %s "
             "Include Dirs: %s. Exclude Dirs: %s", src_dir,
             str(include)[1:-1], str(exclude)[1:-1],
             str(include_dirs)[1:-1], str(exclude_dirs)[1:-1]
             )

    if not os.path.isdir(src_dir):
        LOG.warning("Directory %s does not exist", src_dir)
        return files

    for root, folders, local_files in os.walk(src_dir):
        if not recurse and root != src_dir:
            continue
        root_path = root.split('/')
        excluded = False
        for path in root_path:
            if _matches_one_of(path, exclude_dirs):
                LOG.debug("Directory %s matched an exclude, skipping", root)
                excluded = True
                break
        if excluded:
            continue
        LOG.debug("Scanning '%s' - Folders: %s Files: %s", root, folders,
                  local_files)
        files += [f"{root}/{file}" for file in _filter_files(local_files,
                                                             include, exclude)]
    return files



"""
Sharepoint Connector
"""
import json
import os
import logging
import socket

import requests

from .exceptions import APICallFailedSharepointException
from .auth import SharepointAuth
from .url import SharepointURL
from .utils import from_json, strings_from_list, urlencode

DEFAULT_FOLDER = "Shared Documents/"
LOG = logging.getLogger(__name__)


class SharepointConnector:
    """ Sharepoint Connector """
    ACTIONS = {
        'get': requests.get,
        'post': requests.post,
        'put': requests.put
    }
    DEFAULT_CONTENT_TYPE = "multipart/form-data"

    def __init__(self, tenant, client_id, client_secret, site_name=""):
        """
        Initialize SharePoint Connector

        :param tenant: prefix for url
        :param client_id: app client_id for authentication
        :param client_secret: app client_secret for authentication
        :param site_name: name of the site without `sites/` prefix
        """
        self.url = SharepointURL(tenant, site_name)
        self.auth = SharepointAuth(self.url.host, client_id, client_secret)

    def api_call(self, action, url, headers=None, **kwargs):
        """ Generic web request wrapper with authentication """
        headers = headers if headers else {}

        req = self.ACTIONS.get(action.lower(), None)
        if req is None:
            raise Exception(f"Invalid action: {action}")
        content_type = kwargs.pop('content_type', self.DEFAULT_CONTENT_TYPE)
        raise_on_error = kwargs.pop('raise_on_error', False)
        req_headers = self.headers(content_type, headers=headers)
        res = req(url, headers=req_headers, **kwargs)
        if res.status_code >= 400:
            LOG.error("API ERROR: %s: '%s'", res.status_code, res.text)
            if raise_on_error:
                raise APICallFailedSharepointException(res.status_code,
                                                       res.text)
        return res

    def folder_contents(self, folder=DEFAULT_FOLDER):
        """ Get folder contents"""
        folders = self.folders(folder)
        files = self.files(folder)
        contents = folders + files
        return contents

    @property
    def connected(self):
        """ Check if we are connected"""
        try:
            socket.gethostbyname(self.url.host)
            self.api_call('get', f"{self.url.folder('/')}/Folders",
                          raise_on_error=True)
            return True
        except socket.gaierror:
            LOG.error("Invalid host: %s" % self.url.host)
            return False
        except APICallFailedSharepointException as _ex:
            return False

    def folders(self, folder=DEFAULT_FOLDER):
        """ Get list of folders items in a given folder"""
        return self._get_items(folder, "Folder")

    def files(self, folder=DEFAULT_FOLDER):
        """ Get list of files in a given folder"""
        return self._get_items(folder, "File")

    def _get_items(self, folder=DEFAULT_FOLDER, item_type="File"):
        """ Get list of specific type in a given folder (full data)"""
        res = self.api_call('get', f"{self.url.folder(folder)}/{item_type}s")
        data = from_json(res.text)
        data = data.get('d', data)
        data = data.get('results', data)
        return data

    def list_folder_contents(self, folder=DEFAULT_FOLDER, sort=True):
        """ List files and folders in folder as just names"""
        contents = self.list_files(folder) + self.list_folders(folder)
        return sorted(contents) if sort else contents

    def list_folders(self, folder=DEFAULT_FOLDER):
        """ Get list of folder names in a given folder"""
        return self._list_items(folder, "Folder", suffix="/")

    def list_files(self, folder=DEFAULT_FOLDER):
        """ Get list of files in a given folder"""
        return self._list_items(folder, "File")

    def _list_items(self, folder=DEFAULT_FOLDER, item_type="File",
                    **kwargs):
        """ Get a list of items of a specific type (just names)"""
        items = self._get_items(folder, item_type)
        desc = f"{item_type.lower()} from folder '{folder}'"
        return strings_from_list(items, desc=desc, **kwargs)

    def file_exists(self, filename, folder=DEFAULT_FOLDER):
        """
        Check if file exists in a folder

        :param folder:
        :param filename:
        :return: true if file exists
        """
        files = self.list_files(folder)
        return filename in files

    def get_file_info(self, filename, folder=DEFAULT_FOLDER):
        """ get file """
        if not self.file_exists(filename, folder):
            LOG.info("File %s is not in folder %s", filename, folder)
            return False

        url = self.url.file(folder, filename, "")
        LOG.info("Getting file at url %s", url)
        res = self.api_call('get', url)
        if res.status_code < 400:
            return from_json(res.text)
        LOG.error("ERROR: %s: '%s'", res.status_code, res.text)
        return ""

    def get_file(self, filename, folder=DEFAULT_FOLDER):
        """ UNFINISHED get file """
        if not self.file_exists(filename, folder):
            LOG.error("File %s is not in folder %s", filename, folder)
            return False

        folder = urlencode(folder or "")
        filename = urlencode(filename or "")
        headers = self.headers(headers={
            "Accept": "application/octet-stream"
        })
        url = f"{self.url.file(folder, filename)}"
        LOG.info("Getting file at url %s", url)
        res = self.api_call("get", url, headers=headers)
        if res.status_code < 400:
            return res.text
        LOG.error("ERROR: %s: '%s'", res.status_code, res.text)
        return ""

    def check_in_file(self, filename, folder=DEFAULT_FOLDER,
                      comment="Auto Updated"):
        """ Check-in file in folder"""
        url = self.url.file(folder, filename,
                            f"/CheckIn(comment='{comment}',checkintype=0)")
        LOG.info("Checking in file: %s/%s", folder, filename)
        # LOG.info("Checking in URL: %s", url)
        res = self.api_call('post', url)
        return res

    def check_out_file(self, filename, folder=DEFAULT_FOLDER):
        """ Checkout file in folder"""
        url = self.url.file(folder, filename, "/CheckOut()")
        LOG.info("Checking out file: %s/%s", folder, filename)
        # LOG.info("Checking in URL: %s", url)
        res = self.api_call('post', url)
        return res

    def add_folder_path(self, full_folder, known_folders=None):
        """ Add Full folder path as needed """
        if not known_folders:
            known_folders = []
        folder_parts = full_folder.split('/')
        folder = ""
        added = []
        for part in folder_parts:
            folder = f"{folder}/{part}" if folder else part
            if folder in known_folders:
                continue
            self.add_folder(folder)
            added.append(folder)
        return added

    def add_folder(self, folder=DEFAULT_FOLDER):
        """ Ensure folder exists"""
        LOG.info("Adding folder %s", folder)
        data = json.dumps({
            "__metadata": {
                "type": "SP.Folder"
            },
            "ServerRelativeUrl": f"{self.url.site_uri}/{folder}"
        })
        url = f"{self.url.web}/folders"

        LOG.debug("URL: %s", url)
        LOG.debug("data: %s", data)
        res = self.api_call('post', url,
                            data=data,
                            content_type="application/json;odata=verbose",
                            headers={
                                "Content-Length": f"{len(data)}",
                            })
        LOG.debug("%s::%s::%s", res.status_code, res.headers, res.text)
        return res.text

    def upload_file(self, filename, target_file=None, folder=DEFAULT_FOLDER,
                    check_out=True):
        """ Upload file to sharepoint folder """
        if not os.path.isfile(filename):
            LOG.error("Local File %s does not exist, cannot upload", filename)
            return False
        f_info = os.stat(filename)
        if not target_file:
            target_file = os.path.basename(filename)

        if self.file_exists(target_file, folder):
            self.check_in_file(filename=target_file, folder=folder)

        LOG.error("Uploading %s to %s/%s", filename, folder, target_file)

        file_uri = urlencode(
            f"{self.url.site_uri}/{folder}/{target_file}")

        url = self.url.folder(folder, f"/Files/"
                                      f"Add(url='{file_uri}',overwrite=true)")

        headers = self.headers("application/octet-stream", headers={
            "Content-Length": f"{f_info.st_size}"
        })
        # print(f"Headers: \n{to_yaml(headers)}\n---")
        # print(f"url={url}")
        with open(filename, 'rb') as data:
            res = requests.post(url, data=data, headers=headers)

        if res.status_code < 400:
            if check_out:
                self.check_out_file(target_file, folder)
            return res
        LOG.error("ERROR: %s: '%s'", res.status_code, res.text)
        return res

    def upload_page(self, filename, target_file=None, folder=DEFAULT_FOLDER,
                    check_out=True):
        """ Upload file to sharepoint folder """
        if not os.path.isfile(filename):
            LOG.error("Local File %s does not exist, cannot upload", filename)
            return False
        f_info = os.stat(filename)
        if not target_file:
            target_file = os.path.basename(filename)

        self.check_in_file(filename=target_file, folder=folder)

        LOG.error("Uploading %s to %s/%s", filename, folder, target_file)

        file_uri = urlencode(
            f"{self.url.site_uri}/{folder}/{target_file}")

        url = self.url.folder(folder, f"/Files/AddTemplateFile"
                                      f"("
                                      f"urloffile='{file_uri}',"
                                      f"templatefiletype=0"
                                      f")")

        headers = self.headers("application/octet-stream", headers={
            "Content-Length": f"{f_info.st_size}"
        })
        # print(f"Headers: \n{to_yaml(headers)}\n---")
        # print(f"url={url}")
        with open(filename, 'rb') as data:
            res = requests.post(url, data=data, headers=headers)

        if res.status_code < 400:
            if check_out:
                self.check_out_file(target_file, folder)
            return res
        LOG.error("ERROR: %s: '%s'", res.status_code, res.text)
        return res

    def headers(self, content_type=None, headers=None, **more_headers):
        """ Generate request headers"""
        if content_type is None:
            content_type = self.DEFAULT_CONTENT_TYPE
        if headers is None:
            headers = {}

        req_headers = {
            "Authorization": f"Bearer {self.auth.access_token}",
            "Accept": "application/json;odata=verbose",
            "Content-Type": content_type
        }
        return {**req_headers, **more_headers, **headers}



"""
Unit Tests for utils
"""
import os
import unittest

from sp_tool.sharepoint import url


BASE_DIR = os.path.dirname(os.path.dirname(__file__))
TENANT = "tenant"
SITE = "site"


class SharepointURLTests(unittest.TestCase):
    """ Unit Tests for sp_tool.sharepoint.url"""

    def setUp(self):
        """ Setup """
        self.url = url.SharepointURL(TENANT, SITE)
        self.url_no_site = url.SharepointURL(TENANT, None)

    def test_site_uri_site(self):
        """ Check site_uri if site is declared """
        self.assertEqual(self.url.site_uri, f"/sites/{SITE}")

    def test_site_uri_no_site(self):
        """ Check site_uri if site is not declared """
        self.assertEqual(self.url_no_site.site_uri, "/")

    def test_host_site(self):
        """ test url.host"""
        self.assertEqual(self.url.host, f"{TENANT}.sharepoint.com")

    def test_host_no_site(self):
        """ test url.host"""
        self.assertEqual(self.url_no_site.host, f"{TENANT}.sharepoint.com")


if __name__ == '__main__':
    unittest.main()



"""
Main CLI entry point for sp-tool
"""
import logging
from argparse import Namespace

import click

from sp_tool.sharepoint.utils import to_yaml
from sp_tool.tool.sp_tool import SharepointTool, DEFAULT_CONFIG

from sp_tool.tool.logging import initialize_logging, LOG

__created__ = "09-08-2021"
__updated__ = "10-25-2021"
__version__ = '1.0.1'
ENV_PREFIX = 'SP_TOOL'
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
DEF = Namespace(**DEFAULT_CONFIG)


@click.group(subcommand_metavar='ACTION',
             context_settings=CONTEXT_SETTINGS,
             epilog="""

Have fun!
             """)
@click.pass_context
@click.version_option(version=f"{__version__}({__updated__})")
@click.option('--dry-run/--no-dry-run', default=DEF.dry_run, show_default=True,
              help='If set, do not actually upload')
@click.option('--debug', is_flag=True, help='Debug mode', show_default=True)
@click.option('--site', default="", help='Number of greetings.',
              show_default=True)
@click.option('--tenant', prompt='Tenant Id?', help='Sharepoint Tenant id',
              show_default=True)
@click.option('--recurse/--no-recurse', default=DEF.recurse, show_default=True,
              help='If true, recurse into subdirectories')
@click.option('--checkout/--no-checkout', default=DEF.checkout,
              show_default=True,
              help='If true, recurse into subdirectories')
@click.option('--base-path', default=DEF.base_path, show_default=True,
              help='Base for uploading to sharepoint')
@click.option('--exclude', default=DEF.exclude, multiple=True,
              help='File name(s) to exclude when scanning for files to upload')
@click.option('--exclude-dirs', default=DEF.exclude_dirs, multiple=True,
              help='Directory name(s) to exclude when scanning for files to '
                   'upload(only relevant when recurse is on)')
@click.option('--include', default=DEF.include, multiple=True,
              help='File name(s) to include when scanning for files to upload')
@click.option('--include-dirs', default=DEF.include_dirs, multiple=True,
              help='Directory name(s) to include when scanning for files to '
                   'upload(only relevant when recurse is on)')
@click.option('--path', default=DEF.path, show_default=True,
              help='Path relative to base for uploading to sharepoint')
@click.option('--client-id', default=DEF.client_id, help='Sharepoint Client ID')
@click.option('--secret', default=DEF.secret, help='Sharepoint Secret Token')
def cli(ctx, debug, **kwargs):
    """
        Sharepoint API interface tool for publushing to Microsoft-hosted
        Sharepoint Instances.

        Typical use is with publish command:

            sp-tool [OPTIONS] publish ./dir

    """
    ctx.ensure_object(dict)
    ctx.obj.update(kwargs)
    ctx.obj['debug'] = debug
    if debug:
        LOG.setLevel(logging.DEBUG)
        LOG.debug("Debug Level is On")


@cli.command('register')
@click.pass_context
def register(ctx, **kwargs):
    """ Show effective config and exit"""
    ctx.ensure_object(dict)
    opts = {**DEFAULT_CONFIG, **ctx.obj, **kwargs}
    opts = Namespace(**opts)
    baseurl = "https://%s.sharepoint.com%s" % (
        opts.tenant, f"/sites/{opts.site}" if opts.site else "")

    print(f"""
How to register your app so that you have a client id and secret:
=================================================================

We cannot yet auto-register for you, but we can give you the urls
you will need based on inputs:

Step 1:
=======
Go to:  {baseurl}/_layouts/15/appregnew.aspx

Register your app. Record generated client id and token. Use
whatever values you want for Title/App Domain/Url - they are not
really used.

Step 2:
=======
Go to:  {baseurl}/_layouts/15/appinv.aspx

Enter the client id and do a lookup

Fill in permissions (see docs)

Click on Create and Then Trust.

""")


@cli.command('publish', short_help='Publish to SharePoint')
@click.argument('source_dir', nargs=1, required=True,
                type=click.Path(file_okay=False, exists=True))
@click.pass_context
def publish(ctx, **kwargs):
    """ Publish SOURCE_DIR to SharePoint"""
    ctx.ensure_object(dict)
    opts = {**ctx.obj, **kwargs}
    if not opts.get("tenant"):
        LOG.error("ERROR: Must provide tenant parameter")
        return -1
    if opts.get("secret") and opts.get("client_id"):
        LOG.info("Publishing to sharepoint")
        LOG.debug("Effective Config: \n---\n%s",
                  to_yaml(_normalize_opts(**opts)))
        tool = SharepointTool(**opts)
        return tool.publish()
    register.invoke(ctx)
    LOG.error("ERROR: You must provide client id and a secret")
    return -2


def _normalize_opts(**kwargs):
    """ Normalize/Mask options for safe printing"""
    opts = dict(**kwargs)
    secret = opts.get('secret') or ''
    secret_len = len(secret)
    if secret_len > 6:
        opts['secret'] = secret[0:3] + '*' * (secret_len - 6) + secret[-3:]
    return opts


@cli.command('config')
@click.pass_context
def config(ctx, **kwargs):
    """ Show effective config and exit"""
    ctx.ensure_object(dict)
    opts = {**DEFAULT_CONFIG, **ctx.obj, **kwargs}
    LOG.info("Effective Config: \n---\n%s", to_yaml(_normalize_opts(**opts)))


def main():
    """ Main Entry point """
    initialize_logging()
    cli(auto_envvar_prefix=ENV_PREFIX)  # pylint: disable=no-value-for-parameter


if __name__ == '__main__':
    main()



"""
Sharepoint utils
"""
import json
import urllib.parse
import logging
import yaml

LOG = logging.getLogger(__name__)


def urlencode(string):
    """ Encode string to URL"""
    return urllib.parse.quote(string or "")


def to_yaml(data):
    """ Dump data to yaml string """
    return yaml.safe_dump(data, default_flow_style=False)


def from_json(json_text):
    """ Decode from JSON from text"""
    return json.loads(json_text)


def strings_from_list(item_list, desc="item", field="Name", **kwargs):
    """ Convert a list of items into a list of strings by field """
    sort = bool(kwargs.get('sort', True))
    prefix = kwargs.get('prefix', "")
    suffix = kwargs.get('suffix', "")

    items = []
    for item in item_list:
        if field in item:
            value = item.get(field, None)
            items.append(None if value is None else f"{prefix}{value}{suffix}")
        else:
            LOG.info("Skipping invalid %s : %s", desc, item)
    return sorted(items) if sort else items



"""
Sharepoint Authenticator

"""
import functools
import json

import requests


class SharepointAuth:
    """
    Sharepoint Authenticator

    Authenticates with Sharepoint and gets auth token
    """

    def __init__(self, host, client_id, client_secret):
        """
        Initialize Authenticator

        :param host: site host
        :param client_id: authentication client id
        :param client_secret: authentication secret
        """
        self.host = host
        self.client_id = client_id
        self.client_secret = client_secret

    @property
    @functools.lru_cache(maxsize=None)
    def _remote_data(self):
        """ Get login data"""
        res = requests.get(f"https://{self.host}/_vti_bin/client.svc/",
                           headers={'Authorization': 'Bearer'})
        resp = {}
        for item in res.headers.get('WWW-Authenticate', '').split(','):
            parts = item.split('=', 1)
            if len(parts) > 1:
                resp[parts[0]] = parts[1]
        return (resp.get('Bearer realm', None).strip('"'),
                resp.get('client_id', None).strip('"'))

    @property
    @functools.lru_cache(maxsize=None)
    def bearer_realm(self):
        """ Get token Id"""
        bearer_realm, _ = self._remote_data
        return bearer_realm

    @property
    @functools.lru_cache(maxsize=None)
    def app_id(self):
        """ Get token Id"""
        _, app_id = self._remote_data
        return app_id

    @property
    @functools.lru_cache(maxsize=None)
    def resource_id(self):
        """ get resource id"""
        return f"{self.app_id}/{self.host}@{self.bearer_realm}"

    @property
    @functools.lru_cache(maxsize=None)
    def auth_url(self):
        """ Get auth url"""
        return f"https://accounts.accesscontrol.windows.net/" \
               f"{self.bearer_realm}/tokens/OAuth/2"

    @property
    def full_client_id(self):
        """ Get qualified client id"""
        return f"{self.client_id}@{self.bearer_realm}"

    @property
    @functools.lru_cache(maxsize=None)
    def bearer_token(self):
        """ Get bearer token """
        data = {"client_id": self.full_client_id,
                "client_secret": self.client_secret,
                "resource": self.resource_id,
                "grant_type": "client_credentials"}
        headers = {"Content-type": "application/x-www-form-urlencoded"}
        res = requests.post(self.auth_url, data=data, headers=headers)

        bearer_token = json.loads(res.text)
        return bearer_token

    @property
    def access_token(self):
        """ get bearer access token"""
        return self.bearer_token.get("access_token", None)

    @property
    def connected(self):
        """
        Check if we are connected and have access token

        :return: true if we have access token
        """
        return self.access_token is not None



#!/usr/bin/env python3
import sys

from sp_tool.sharepoint import SharepointConnector

TENANT = "mycompany"
SITE = "MyTeam"
# Dummy Credentials
CLIENT_ID = "db87c4c8-9a1a-4b53-8aa7-278fa558af34"
SECRET = "6FuTGk52OPCaSJAlvsJ/AhTsjjZlGpwLqKA2BCx873E="


def main():
    """ Simple Example"""
    location = sys.argv[1] if len(sys.argv) > 1 else "/"

    # Create Client
    sharepoint = SharepointConnector(
        tenant=TENANT,
        client_id=CLIENT_ID,
        client_secret=SECRET,
        site_name=SITE
    )

    # Test Connection
    if sharepoint.connected:
        # list all folders and files in default directory
        items = sharepoint.list_folder_contents(location)
        print(f"Found {len(items)} items in {location}")
        for item in items:
            print(f"Found: {item}")
    else:
        print("Failed to connect to sharepoint")


if __name__ == "__main__":
    main()



"""
Note that bulk of the configuation is in the package.cfg file
"""
try:
    import ConfigParser as configparser
except ImportError:
    import configparser

import os.path
from pprint import pprint

from setuptools import setup, find_packages


BASEDIR = os.path.dirname(__file__)

requirements = []
with open(os.path.join(BASEDIR, 'requirements.txt'), "r") as f:
    requirements = f.readlines()

with open("README.md", "r") as fh:
    long_description = fh.read()


# Read Package info from a CONFIG file package.cfg
CONFIG = configparser.ConfigParser({})
CONFIG.add_section("Package")
CONFIG.add_section("FindPackages")
CONFIG.read(os.path.join(BASEDIR, 'package.cfg'))
PKG_INFO = dict(CONFIG.items('Package'))

FIND_PKGS = dict(CONFIG.items('FindPackages'))

for item in ['include', 'exclude']:
    if item in FIND_PKGS:
        FIND_PKGS[item] = FIND_PKGS[item].split(',')

PKG_INFO.update({
    'python_requires': '>=3.6',
    'packages': find_packages(**(FIND_PKGS)),
    'package_dir': {'': 'src'},
    'package_data': {'': ['/package.cfg*']},
    'install_requires': [req.strip() for req in requirements if not req.startswith('#')],
    'long_description': long_description,
    'long_description_content_type': "text/markdown",
    'classifiers': [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    'entry_points': {
        'console_scripts': [
            'sp-tool = sp_tool.tool.main:main',
        ],
    }
})

setup(**PKG_INFO)
